import functools
from inspect import isclass

from pystachio.naming import Ref, Namable, frozendict

class TypeCheck(object):
  """
    Encapsulate the results of a type check pass.
  """
  class Error(Exception):
    pass

  @staticmethod
  def success():
    return TypeCheck(True, "")

  @staticmethod
  def failure(msg):
    return TypeCheck(False, msg)

  def __init__(self, success, message):
    self._success = success
    self._message = message

  def message(self):
    return self._message

  def ok(self):
    return self._success

  def __repr__(self):
    if self.ok():
      return 'TypeCheck(OK)'
    else:
      return 'TypeCheck(FAILED): %s' % self._message


class TypeFactoryType(type):
  _TYPE_FACTORIES = {}

  def __new__(mcs, name, parents, attributes):
    if 'PROVIDES' not in attributes:
      return type.__new__(mcs, name, parents, attributes)
    else:
      provides = attributes['PROVIDES']
      new_type = type.__new__(mcs, name, parents, attributes)
      TypeFactoryType._TYPE_FACTORIES[provides] = new_type
      return new_type


TypeFactoryClass = TypeFactoryType('TypeFactoryClass', (object,), {})
class TypeFactory(TypeFactoryClass):
  @staticmethod
  def get_factory(type_name):
    assert type_name in TypeFactoryType._TYPE_FACTORIES, (
      'Unknown type: %s, Existing factories: %s' % (
        type_name, TypeFactoryType._TYPE_FACTORIES.keys()))
    return TypeFactoryType._TYPE_FACTORIES[type_name]

  @staticmethod
  def create(type_dict, *type_parameters):
    """
      Implemented by the TypeFactory to produce a new type.

      Should return:
        reified type
        (with usable type.__name__)
    """
    raise NotImplementedError("create unimplemented for: %s" % repr(type_parameters))

  @staticmethod
  def new(type_dict, type_factory, *type_parameters):
    """
      Create a fully reified type from a type schema.
    """
    type_tuple = (type_factory,) + type_parameters
    if type_tuple not in type_dict:
      factory = TypeFactory.get_factory(type_factory)
      reified_type = factory.create(type_dict, *type_parameters)
      type_dict[type_tuple] = reified_type
    return type_dict[type_tuple]

  @staticmethod
  def wrapper(factory):
    assert issubclass(factory, TypeFactory)
    def wrapper_function(*type_parameters):
      return TypeFactory.new({}, factory.PROVIDES, *tuple(
        [typ.serialize_type() for typ in type_parameters]))
    return wrapper_function

  @staticmethod
  def load(type_tuple, into=None):
    """
      Determine all types touched by loading the type and deposit them into
      the particular namespace.
    """
    type_dict = {}
    TypeFactory.new(type_dict, *type_tuple)
    deposit = into if (into is not None and isinstance(into, dict)) else {}
    for reified_type in type_dict.values():
      deposit[reified_type.__name__] = reified_type
    return deposit

  @staticmethod
  def load_json(json_list, into=None):
    """
      Determine all types touched by loading the type and deposit them into
      the particular namespace.
    """
    def l2t(obj):
      if isinstance(obj, list):
        return tuple(l2t(L) for L in obj)
      elif isinstance(obj, dict):
        return frozendict(obj)
      else:
        return obj
    return TypeFactory.load(l2t(json_list), into=into)

  @staticmethod
  def load_file(filename, into=None):
    import json
    with open(filename) as fp:
      return TypeFactory.load_json(json.load(fp), into=into)


class TypeMetaclass(type):
  def __instancecheck__(cls, other):
    if not hasattr(other, 'type_parameters'):
      return False
    if not hasattr(other, '__class__'):
      return False
    if cls.__name__ != other.__class__.__name__:
      return False
    return cls.type_factory() == other.type_factory() and (
      cls.type_parameters() == other.type_parameters())

  def __new__(mcls, name, parents, attributes):
    return type.__new__(mcls, name, parents, attributes)


class Type(object):
  @classmethod
  def type_factory(cls):
    """ Return the name of the factory that produced this class. """
    raise NotImplementedError

  @classmethod
  def type_parameters(cls):
    """ Return the type parameters used to produce this class. """
    raise NotImplementedError

  @classmethod
  def serialize_type(cls):
    return (cls.type_factory(),) + cls.type_parameters()

  @classmethod
  def dump(cls, fp):
    import json
    json.dump(cls.serialize_type(), fp)

  def check(self):
    """
      Returns a TypeCheck object explaining whether or not a particular
      instance of this object typechecks.
    """
    raise NotImplementedError


class TypeEnvironment(object):
  @staticmethod
  def deserialize(klazz_bindings, type_dict):
    unbound_types = []
    bound_types = {}

    for binding in klazz_bindings:
      if len(binding) == 1:
        unbound_types.append(TypeFactory.new(type_dict, *binding[0]))
      elif len(binding) == 2:
        bound_types[binding[1]] = TypeFactory.new(type_dict, *binding[0])
      else:
        raise ValueError('Expected 1- or 2-tuple to TypeEnvironment.deserialize')

    return TypeEnvironment(*unbound_types, **bound_types)

  def __init__(self, *types, **bound_types):
    for typ in types + tuple(bound_types.values()):
      if not isclass(typ) or not issubclass(typ, Namable):
        raise TypeError('Type annotations must be subtypes of Namable, got %s instead!' % repr(typ))
    self._unbound_types = types
    self._bound_types = bound_types

  def merge(self, other):
    self_serialized = self.serialize()
    other_serialized = other.serialize()
    combined = set(self_serialized + other_serialized)
    return TypeEnvironment.deserialize(tuple(combined), {})

  def serialize(self):
    serialized_bindings = []
    for typ in self._unbound_types:
      serialized_bindings.append((typ.serialize_type(),))
    for (name, typ) in self._bound_types.items():
      serialized_bindings.append((typ.serialize_type(), name))
    return tuple(serialized_bindings)

  def covers(self, ref):
    """
      Does this TypeEnvironment cover the ref?
    """
    for binding in self._unbound_types:
      if binding.provides(ref):
        return True
    for bound_name, binding in self._bound_types.items():
      scoped_ref = Ref.from_address(bound_name).scoped_to(ref)
      if scoped_ref is not None and binding.provides(scoped_ref):
        return True
    return False

  def __str__(self):
    return 'TypeEnvironment(%s, %s)' % (
      ' '.join(unbound.__name__ for unbound in self._unbound_types),
      ' '.join('%s=>%s' % (name, bound.__name__) for name, bound in self._bound_types.items()))



from collections import Iterable, Mapping, Sequence
import copy
from inspect import isclass

from pystachio.base import Object
from pystachio.compatibility import Compatibility
from pystachio.naming import Namable, Ref, frozendict
from pystachio.typing import (
  Type,
  TypeCheck,
  TypeEnvironment,
  TypeFactory,
  TypeMetaclass)

class ListFactory(TypeFactory):
  PROVIDES = 'List'

  @staticmethod
  def create(type_dict, *type_parameters):
    """
      Construct a List containing type 'klazz'.
    """
    assert len(type_parameters) == 1
    klazz = TypeFactory.new(type_dict, *type_parameters[0])
    assert isclass(klazz)
    assert issubclass(klazz, Object)
    return TypeMetaclass('%sList' % klazz.__name__, (ListContainer,), { 'TYPE': klazz })


class ListContainer(Namable, Object, Type):
  """
    The List container type.  This is the base class for all user-generated
    List types.  It won't function as-is, since it requires cls.TYPE to be
    set to the contained type.  If you want a concrete List type, see the
    List() function.
  """
  def __init__(self, vals):
    self._values = self._coerce_values(copy.copy(vals))
    Object.__init__(self)

  def get(self):
    return tuple([v.get() for v in self._values])

  def copy(self):
    new_self = self.__class__(self._values)
    new_self._scopes = copy.copy(self.scopes())
    new_self._modulo = copy.copy(self.modulo())
    return new_self

  def __hash__(self):
    return hash(self.get())

  def __repr__(self):
    si, _ = self.interpolate()
    return '%s(%s)' % (self.__class__.__name__,
      ', '.join(str(v) if Compatibility.PY3 else unicode(v) for v in si._values))

  def __iter__(self):
    si, _ = self.interpolate()
    return iter(si._values)

  def __getitem__(self, index_or_slice):
    si, _ = self.interpolate()
    return si._values[index_or_slice]

  def __contains__(self, item):
    si, _ = self.interpolate()
    if isinstance(item, self.TYPE):
      return item in si._values
    else:
      return item in si.get()

  def __eq__(self, other):
    if not isinstance(other, ListContainer): return False
    if self.TYPE.serialize_type() != other.TYPE.serialize_type(): return False
    si, _ = self.interpolate()
    oi, _ = other.interpolate()
    return si._values == oi._values

  @staticmethod
  def isiterable(values):
    return isinstance(values, Sequence) and not isinstance(values, Compatibility.stringy)

  def _coerce_values(self, values):
    if not ListContainer.isiterable(values):
      raise ValueError("ListContainer expects an iterable, got %s" % repr(values))
    def coerced(value):
      return value if isinstance(value, self.TYPE) else self.TYPE(value)
    return tuple([coerced(v) for v in values])

  def check(self):
    assert ListContainer.isiterable(self._values)
    for element in self._values:
      assert isinstance(element, self.TYPE)
      typecheck = element.in_scope(*self.scopes()).provided(self.modulo()).check()
      if not typecheck.ok():
        return TypeCheck.failure("Element in %s failed check: %s" % (self.__class__.__name__,
          typecheck.message()))
    return TypeCheck.success()

  def interpolate(self):
    unbound = set()
    interpolated = []
    for element in self._values:
      einterp, eunbound = element.in_scope(*self.scopes()).provided(self.modulo()).interpolate()
      interpolated.append(einterp)
      unbound.update(eunbound)
    return self.__class__(interpolated), list(unbound)

  @classmethod
  def provides(cls, ref):
    assert isinstance(ref, Ref)
    if isinstance(ref.action(), ref.Index):
      if ref.rest().is_empty():
        return True
      else:
        if issubclass(cls.TYPE, Namable):
          return cls.TYPE.provides(ref.rest())
    return False

  def find(self, ref):
    if not ref.is_index():
      raise Namable.NamingError(self, ref)
    try:
      intvalue = int(ref.action().value)
    except ValueError:
      raise Namable.NamingError(self, ref)
    if len(self._values) <= intvalue:
      raise Namable.NotFound(self, ref)
    else:
      namable = self._values[intvalue]
      if ref.rest().is_empty():
        return namable.in_scope(*self.scopes())
      else:
        if not isinstance(namable, Namable):
          raise Namable.Unnamable(namable)
        else:
          return namable.in_scope(*self.scopes()).find(ref.rest())

  @classmethod
  def type_factory(cls):
    return 'List'

  @classmethod
  def type_parameters(cls):
    return (cls.TYPE.serialize_type(),)

List = TypeFactory.wrapper(ListFactory)


class MapFactory(TypeFactory):
  PROVIDES = 'Map'

  @staticmethod
  def create(type_dict, *type_parameters):
    assert len(type_parameters) == 2, 'Type parameters: %s' % repr(type_parameters)
    key_klazz, value_klazz = type_parameters
    key_klazz, value_klazz = (TypeFactory.new(type_dict, *key_klazz),
                              TypeFactory.new(type_dict, *value_klazz))
    assert isclass(key_klazz) and isclass(value_klazz)
    assert issubclass(key_klazz, Object) and issubclass(value_klazz, Object)
    return TypeMetaclass('%s%sMap' % (key_klazz.__name__, value_klazz.__name__), (MapContainer,),
      { 'KEYTYPE': key_klazz, 'VALUETYPE': value_klazz })


class MapContainer(Namable, Object, Type):
  """
    The Map container type.  This is the base class for all user-generated
    Map types.  It won't function as-is, since it requires cls.KEYTYPE and
    cls.VALUETYPE to be set to the appropriate types.  If you want a
    concrete Map type, see the Map() function.

    __init__(dict) => translates to list of tuples & sanity checks
    __init__(tuple) => sanity checks
  """
  def __init__(self, *args):
    """
      Construct a map.

      Input:
        sequence of tuples _or_ a dictionary
    """
    if len(args) == 1 and isinstance(args[0], Mapping):
      self._map = self._coerce_map(copy.copy(args[0]))
    elif all(isinstance(arg, Iterable) and len(arg)==2 for arg in args):
      self._map = self._coerce_tuple(args)
    else:
      raise ValueError("Unexpected input to MapContainer: %s" % repr(args))
    Object.__init__(self)

  def get(self):
    return frozendict((k.get(), v.get()) for (k, v) in self._map)

  def _coerce_wrapper(self, key, value):
    coerced_key = key if isinstance(key, self.KEYTYPE) else self.KEYTYPE(key)
    coerced_value = value if isinstance(value, self.VALUETYPE) else self.VALUETYPE(value)
    return (coerced_key, coerced_value)

  def _coerce_map(self, input_map):
    return tuple(self._coerce_wrapper(key, value) for key, value in input_map.items())

  def _coerce_tuple(self, input_tuple):
    return tuple(self._coerce_wrapper(key, value) for key, value in input_tuple)

  def __hash__(self):
    return hash(self.get())

  def __iter__(self):
    si, _ = self.interpolate()
    return (t[0] for t in si._map)

  def __getitem__(self, key):
    if not isinstance(key, self.KEYTYPE):
      try:
        key = self.KEYTYPE(key)
      except ValueError:
        raise KeyError("%s is not coercable to %s" % self.KEYTYPE.__name__)
    # TODO(wickman) The performance of this should be improved.
    si, _ = self.interpolate()
    for tup in si._map:
      if key == tup[0]:
        return tup[1]
    raise KeyError("%s not found" % key)

  def __contains__(self, item):
    try:
      self[item]
      return True
    except KeyError:
      return False

  def copy(self):
    new_self = self.__class__(*self._map)
    new_self._scopes = copy.copy(self.scopes())
    new_self._modulo = copy.copy(self.modulo())
    return new_self

  def __repr__(self):
    si, _ = self.interpolate()
    return '%s(%s)' % (self.__class__.__name__,
      ', '.join('%s => %s' % (key, val) for key, val in si._map))

  def __eq__(self, other):
    if not isinstance(other, MapContainer): return False
    if self.KEYTYPE.serialize_type() != other.KEYTYPE.serialize_type(): return False
    if self.VALUETYPE.serialize_type() != other.VALUETYPE.serialize_type(): return False
    si, _ = self.interpolate()
    oi, _ = other.interpolate()
    return si._map == oi._map

  def check(self):
    assert isinstance(self._map, tuple)
    for key, value in self._map:
      assert isinstance(key, self.KEYTYPE)
      assert isinstance(value, self.VALUETYPE)
      keycheck = key.in_scope(*self.scopes()).provided(self.modulo()).check()
      valuecheck = value.in_scope(*self.scopes()).provided(self.modulo()).check()
      if not keycheck.ok():
        return TypeCheck.failure("%s key %s failed check: %s" % (self.__class__.__name__,
          key, keycheck.message()))
      if not valuecheck.ok():
        return TypeCheck.failure("%s[%s] value %s failed check: %s" % (self.__class__.__name__,
          key, value, valuecheck.message()))
    return TypeCheck.success()

  def interpolate(self):
    unbound = set()
    interpolated = []
    for key, value in self._map:
      kinterp, kunbound = key.in_scope(*self.scopes()).provided(self.modulo()).interpolate()
      vinterp, vunbound = value.in_scope(*self.scopes()).provided(self.modulo()).interpolate()
      unbound.update(kunbound)
      unbound.update(vunbound)
      interpolated.append((kinterp, vinterp))
    return self.__class__(*interpolated), list(unbound)

  @classmethod
  def provides(cls, ref):
    # TODO(wickman)  Should we be typechecking the ref.action().value?
    assert isinstance(ref, Ref)
    if isinstance(ref.action(), ref.Index):
      if ref.rest().is_empty():
        return True
      else:
        if issubclass(cls.VALUETYPE, Namable):
          return cls.VALUETYPE.provides(ref.rest())
    return False

  def find(self, ref):
    if not ref.is_index():
      raise Namable.NamingError(self, ref)
    kvalue = self.KEYTYPE(ref.action().value)
    for key, namable in self._map:
      if kvalue == key:
        if ref.rest().is_empty():
          return namable.in_scope(*self.scopes())
        else:
          if not isinstance(namable, Namable):
            raise Namable.Unnamable(namable)
          else:
            return namable.in_scope(*self.scopes()).find(ref.rest())
    raise Namable.NotFound(self, ref)

  @classmethod
  def type_factory(cls):
    return 'Map'

  @classmethod
  def type_parameters(cls):
    return (cls.KEYTYPE.serialize_type(), cls.VALUETYPE.serialize_type())


Map = TypeFactory.wrapper(MapFactory)



from sys import version_info
from numbers import (Real, Integral)

class Compatibility(object):
  """2.x + 3.x compatibility"""
  stringy = (str,)
  if version_info[0] == 2:
    stringy = (str, unicode)
  integer = (Integral,)
  real = (Real,)
  numeric = integer + real
  PY2 = version_info[0] == 2
  PY3 = version_info[0] == 3



import copy

from pystachio.base import Object
from pystachio.compatibility import Compatibility
from pystachio.parsing import MustacheParser
from pystachio.typing import Type, TypeFactory, TypeCheck


class SimpleObject(Object, Type):
  """
    A simply-valued (unnamable) object.
  """
  def __init__(self, value):
    self._value = value
    Object.__init__(self)

  def get(self):
    return self._value

  def copy(self):
    new_self = self.__class__(self._value)
    new_self._scopes = copy.copy(self.scopes())
    new_self._modulo = copy.copy(self.modulo())
    return new_self

  def _my_cmp(self, other):
    if self.__class__ != other.__class__:
      return -1
    si, _ = self.interpolate()
    oi, _ = other.interpolate()
    if si._value < oi._value:
      return -1
    elif si._value > oi._value:
      return 1
    else:
      return 0

  def __hash__(self):
    return hash(self._value)

  def __eq__(self, other):
    return self._my_cmp(other) == 0

  def __lt__(self, other):
    return self._my_cmp(other) == -1

  def __gt__(self, other):
    return self._my_cmp(other) == 1

  def __le__(self, other):
    return self._my_cmp(other) <= 0

  def __ge__(self, other):
    return self._my_cmp(other) >= 0

  def __unicode__(self):
    si, _ = self.interpolate()
    return unicode(si._value)

  def __str__(self):
    si, _ = self.interpolate()
    return str(si._value)

  def __repr__(self):
    return '%s(%s)' % (self.__class__.__name__, str(self) if Compatibility.PY3 else unicode(self))

  def interpolate(self):
    if not isinstance(self._value, Compatibility.stringy):
      return self.__class__(self.coerce(self._value)), []
    else:
      joins, unbound = MustacheParser.resolve(self._value, *self.scopes())
      if unbound:
        return self.__class__(joins), [ref for ref in unbound if not self.modulo().covers(ref)]
      else:
        self_copy = self.copy()
        self_copy._value = self_copy.coerce(joins)
        return self_copy, unbound

  @classmethod
  def type_factory(cls):
    return cls.__name__

  @classmethod
  def type_parameters(cls):
    return ()


class String(SimpleObject):
  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, String)
    if isinstance(obj._value, Compatibility.stringy):
      return TypeCheck.success()
    else:
      # TODO(wickman)  Perhaps we should mark uninterpolated Mustache objects as
      # intrinsically non-stringy, because String will never typecheck false given
      # its input constraints.
      return TypeCheck.failure("%s not a string" % repr(obj._value))

  @classmethod
  def coerce(cls, value):
    ACCEPTED_SOURCE_TYPES = Compatibility.stringy + Compatibility.numeric
    if not isinstance(value, ACCEPTED_SOURCE_TYPES):
      raise cls.CoercionError(value, cls)
    return str(value) if Compatibility.PY3 else unicode(value)


class StringFactory(TypeFactory):
  PROVIDES = 'String'
  @staticmethod
  def create(type_dict, *type_parameters):
    return String


class Integer(SimpleObject):
  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, Integer)
    if isinstance(obj._value, Compatibility.integer):
      return TypeCheck.success()
    else:
      return TypeCheck.failure("%s not an integer" % repr(obj._value))

  @classmethod
  def coerce(cls, value):
    ACCEPTED_SOURCE_TYPES = Compatibility.numeric + Compatibility.stringy
    if not isinstance(value, ACCEPTED_SOURCE_TYPES):
      raise cls.CoercionError(value, cls)
    try:
      return int(value)
    except ValueError:
      raise cls.CoercionError(value, cls)


class IntegerFactory(TypeFactory):
  PROVIDES = 'Integer'
  @staticmethod
  def create(type_dict, *type_parameters):
    return Integer


class Float(SimpleObject):
  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, Float)
    if isinstance(obj._value, Compatibility.real + Compatibility.integer):
      return TypeCheck.success()
    else:
      return TypeCheck.failure("%s not a float" % repr(obj._value))

  @classmethod
  def coerce(cls, value):
    ACCEPTED_SOURCE_TYPES = Compatibility.numeric + Compatibility.stringy
    if not isinstance(value, ACCEPTED_SOURCE_TYPES):
      raise cls.CoercionError(value, cls)
    try:
      return float(value)
    except ValueError:
      raise cls.CoercionError(value, cls)

class FloatFactory(TypeFactory):
  PROVIDES = 'Float'
  @staticmethod
  def create(type_dict, *type_parameters):
    return Float



from collections import Mapping
import copy
from inspect import isclass
import json

from pystachio.base import Object, Environment
from pystachio.naming import Ref, Namable, frozendict
from pystachio.typing import (
  Type,
  TypeCheck,
  TypeEnvironment,
  TypeFactory,
  TypeMetaclass)

class Empty(object):
  """The Empty sentinel representing an unspecified field."""
  pass

class TypeSignature(object):
  """
    Type metadata for composite type schemas.
  """

  def __init__(self, cls, required=False, default=Empty):
    assert isclass(cls)
    assert issubclass(cls, Object)
    if default is not Empty and not isinstance(default, cls):
      self._default = cls(default)
    else:
      self._default = default
    self._cls = cls
    self._required = required

  def serialize(self):
    return (self.required,
            self.default.get() if not self.empty else (),
            self.empty,
            self.klazz.serialize_type())

  @staticmethod
  def deserialize(sig, type_dict):
    req, default, empty, klazz_schema = sig
    real_class = TypeFactory.new(type_dict, *klazz_schema)
    if not empty:
      return TypeSignature(real_class, default=real_class(default), required=req)
    else:
      return TypeSignature(real_class, required=req)

  def __eq__(self, other):
    return (self.klazz.serialize_type() == other.klazz.serialize_type() and
            self.required == other.required and
            self.default == other.default and
            self.empty == other.empty)

  def __ne__(self, other):
    return not (self == other)

  def __repr__(self):
    return 'TypeSignature(%s, required: %s, default: %s, empty: %s)' % (
      self.klazz.__name__, self.required, self.default, self.empty)

  @property
  def klazz(self):
    return self._cls

  @property
  def required(self):
    return self._required

  @property
  def default(self):
    return self._default

  @property
  def empty(self):
    return self._default is Empty

  @staticmethod
  def wrap(sig):
    if isclass(sig) and issubclass(sig, Object):
      return TypeSignature(sig)
    elif isinstance(sig, TypeSignature):
      return sig


def Required(cls):
  """
    Helper to make composite types read succintly.  Wrap a type and make its
    specification required during type-checking of composite types.
  """
  return TypeSignature(cls, required=True)


def Default(cls, default):
  """
    Helper to make composite types read succintly.  Wrap a type and assign it a
    default if it is unspecified in the construction of the composite type.
  """
  return TypeSignature(cls, required=False, default=default)


def Provided(*types, **bound_types):
  """
    Specify the requirement that this Struct is only interpolated within the context
    of certain other Structs.

    e.g.

    class JobEnvironment(Struct):
      user  = String
      group = String
      ram   = Integer

    @Provided(job = JobEnvironment)
    class Job(Struct):
      name    = String
      command = Default(String, 'echo {{job.user}} from {{job.group}}!')
  """
  def annotator(klazz):
    assert issubclass(klazz, Structural)
    klazz_factory, klazz_name, klazz_bindings, klazz_attributes = klazz.serialize_type()

    old_type_environment = TypeEnvironment.deserialize(klazz_bindings, {})
    new_type_environment = TypeEnvironment(*types, **bound_types)
    return TypeFactory.new({}, 'Struct',
      klazz_name,
      new_type_environment.merge(old_type_environment).serialize(),
      klazz_attributes)
  return annotator


class StructFactory(TypeFactory):
  PROVIDES = 'Struct'

  @staticmethod
  def create(type_dict, *type_parameters):
    """
      StructFactory.create(*type_parameters) expects:

        class name,
        ((binding requirement1,),
          (binding requirement2, bound_to_scope),
           ...),
        ((attribute_name1, attribute_sig1 (serialized)),
         (attribute_name2, attribute_sig2 ...),
         ...
         (attribute_nameN, ...))
    """
    name, bindings, parameters = type_parameters
    for param in parameters:
      assert isinstance(param, tuple)
    typemap = dict((attr, TypeSignature.deserialize(param, type_dict))
                   for attr, param in parameters)
    bindings = TypeEnvironment.deserialize(bindings, type_dict)
    attributes = {'TYPEMAP': typemap, 'REQUIRES': bindings}
    return TypeMetaclass(str(name), (Structural,), attributes)


class StructMetaclass(type):
  """
    Schema-extracting metaclass for Struct objects.
  """
  @staticmethod
  def attributes_to_parameters(attributes):
    parameters = []
    for attr_name, attr_value in attributes.items():
      sig = TypeSignature.wrap(attr_value)
      if sig:
        parameters.append((attr_name, sig.serialize()))
    return tuple(parameters)

  def __new__(mcs, name, parents, attributes):
    if any(parent.__name__ == 'Struct' for parent in parents):
      type_parameters = StructMetaclass.attributes_to_parameters(attributes)
      return TypeFactory.new({}, 'Struct', name, (), type_parameters)
    else:
      return type.__new__(mcs, name, parents, attributes)


StructMetaclassWrapper = StructMetaclass('StructMetaclassWrapper', (object,), {})

class Structural(Object, Type, Namable):
  """A Structural base type for composite objects."""

  def __init__(self, *args, **kw):
    self._init_schema_data()
    for arg in args:
      if not isinstance(arg, Mapping):
        raise ValueError('Expected dictionary argument, got %s' % repr(arg))
      self._update_schema_data(**arg)
    self._update_schema_data(**copy.copy(kw))
    Object.__init__(self)

  def get(self):
    return frozendict((k, v.get()) for k, v in self._schema_data.items() if v is not Empty)

  def _init_schema_data(self):
    self._schema_data = {}
    for attr in self.TYPEMAP:
      self._schema_data[attr] = self.TYPEMAP[attr].default
    self._schema_data = frozendict(self._schema_data)

  def _process_schema_attribute(self, attr, value):
    if attr not in self.TYPEMAP:
      raise AttributeError('Unknown schema attribute %s' % attr)
    schema_type = self.TYPEMAP[attr]
    if value is Empty:
      return Empty
    elif isinstance(value, schema_type.klazz):
      return value
    else:
      return schema_type.klazz(value)

  def _update_schema_data(self, **kw):
    for attr, value in kw.items():
      self._schema_data[attr] = self._process_schema_attribute(attr, value)

  def copy(self):
    new_object = self.__class__(**self._schema_data)
    new_object._scopes = copy.copy(self.scopes())
    new_object._modulo = copy.copy(self.modulo())
    return new_object

  def __call__(self, **kw):
    new_self = self.copy()
    new_self._update_schema_data(**copy.copy(kw))
    return new_self

  def __eq__(self, other):
    if not isinstance(other, Structural): return False
    if self.TYPEMAP != other.TYPEMAP: return False
    si = self.interpolate()
    oi = other.interpolate()
    return si[0]._schema_data == oi[0]._schema_data

  def __repr__(self):
    si, _ = self.interpolate()
    return '%s(%s)' % (
      self.__class__.__name__,
      ', '.join('%s=%s' % (key, val) for key, val in si._schema_data.items() if val is not Empty)
    )

  def __getattr__(self, attr):
    if not hasattr(self, 'TYPEMAP'):
      raise AttributeError

    if attr.startswith('has_'):
      if attr[4:] in self.TYPEMAP:
        return lambda: self._schema_data[attr[4:]] != Empty

    if attr not in self.TYPEMAP:
      raise AttributeError("%s has no attribute %s" % (self.__class__.__name__, attr))

    return lambda: self.interpolate_key(attr)

  def check(self):
    for name, signature in self.TYPEMAP.items():
      if self._schema_data[name] is Empty and signature.required:
        return TypeCheck.failure('%s[%s] is required.' % (self.__class__.__name__, name))
      elif self._schema_data[name] is not Empty:
        type_check = (self._schema_data[name].in_scope(*self.scopes())
                                             .provided(self.modulo())
                                             .check())
        if type_check.ok():
          continue
        else:
          return TypeCheck.failure('%s[%s] failed: %s' % (self.__class__.__name__, name,
            type_check.message()))
    return TypeCheck.success()

  def modulo(self):
    return super(Structural, self).modulo().merge(self.REQUIRES)

  def _self_scope(self):
    return dict((key, value) for (key, value) in self._schema_data.items()
                if value is not Empty)

  def scopes(self):
    return [Environment(self._self_scope())] + self._scopes

  def interpolate(self):
    unbound = set()
    interpolated_schema_data = {}
    scopes = self.scopes()
    modulo = self.modulo()
    for key, value in self._schema_data.items():
      if value is Empty:
        interpolated_schema_data[key] = Empty
      else:
        vinterp, vunbound = value.in_scope(*scopes).provided(modulo).interpolate()
        unbound.update(vunbound)
        interpolated_schema_data[key] = vinterp
    return self.__class__(**interpolated_schema_data), list(unbound)

  def interpolate_key(self, attribute):
    if self._schema_data[attribute] is Empty:
      return Empty
    vinterp, _ = (
      self._schema_data[attribute].in_scope(*self.scopes())
                                  .provided(self.modulo())
                                  .interpolate())
    return self._process_schema_attribute(attribute, vinterp)

  @classmethod
  def type_factory(cls):
    return 'Struct'

  @classmethod
  def type_parameters(cls):
    attrs = []
    if hasattr(cls, 'TYPEMAP'):
      attrs = sorted([(attr, sig.serialize()) for attr, sig in cls.TYPEMAP.items()])
    bindings = cls.REQUIRES.serialize()
    return (cls.__name__, bindings, tuple(attrs))

  @classmethod
  def provides(cls, ref):
    assert isinstance(ref, Ref)
    if isinstance(ref.action(), Ref.Dereference):
      if ref.action().value in cls.TYPEMAP:
        if ref.rest().is_empty():
          return True
        else:
          klazz = cls.TYPEMAP[ref.action().value].klazz
          if issubclass(klazz, Namable):
            return cls.TYPEMAP[ref.action().value].klazz.provides(ref.rest())
    return False

  @classmethod
  def _filter_against_schema(cls, values):
    return dict((key, val) for (key, val) in values.items()
                if key in cls.TYPEMAP)

  @classmethod
  def json_load(cls, fp, strict=False):
    return cls(json.load(fp) if strict else cls._filter_against_schema(json.load(fp)))

  @classmethod
  def json_loads(cls, json_string, strict=False):
    return cls(json.loads(json_string) if strict
               else cls._filter_against_schema(json.loads(json_string)))

  def json_dump(self, fp):
    d, _ = self.interpolate()
    return json.dump(d.get(), fp)

  def json_dumps(self):
    d, _ = self.interpolate()
    return json.dumps(d.get())

  def find(self, ref):
    if not ref.is_dereference():
      raise Namable.NamingError(self, ref)
    name = ref.action().value
    if name not in self.TYPEMAP or self._schema_data[name] is Empty:
      raise Namable.NotFound(self, ref)
    else:
      namable = self._schema_data[name]
      if ref.rest().is_empty():
        return namable.in_scope(*self.scopes())
      else:
        if not isinstance(namable, Namable):
          raise Namable.Unnamable(namable)
        else:
          return namable.in_scope(*self.scopes()).find(ref.rest())


class Struct(StructMetaclassWrapper, Structural):
  """
    Schema-based composite objects, e.g.

      class Employee(Struct):
        first = Required(String)
        last  = Required(String)
        email = Required(String)
        phone = String

      Employee(first = "brian", last = "wickman", email = "wickman@twitter.com").check()

    They're purely functional data structures and behave more like functors.
    In other words they're immutable:

      >>> brian = Employee(first = "brian")
      >>> brian(last = "wickman")
      Employee(last=String(wickman), first=String(brian))
      >>> brian
      Employee(first=String(brian))
  """
  pass



from pprint import pformat

from pystachio.compatibility import Compatibility
from pystachio.naming import (
  Ref,
  Namable)
from pystachio.parsing import MustacheParser
from pystachio.typing import (
  TypeCheck,
  TypeEnvironment)


class Environment(Namable):
  """
    A mount table for Refs pointing to Objects or arbitrary string substitutions.
  """

  @staticmethod
  def wrap(value):
    if isinstance(value, dict):
      return Environment(value)
    elif isinstance(value, (Environment, Object)):
      return value
    else:
      if isinstance(value, Compatibility.numeric + Compatibility.stringy):
        return str(value)
      else:
        raise ValueError('Error in Environment.wrap(%s)' % repr(value))

  def _assimilate_dictionary(self, d):
    for key, val in d.items():
      val = Environment.wrap(val)
      rkey = Ref.wrap(key)
      if isinstance(val, Environment):
        for vkey, vval in val._table.items():
          self._table[rkey + vkey] = vval
      else:
        self._table[rkey] = val

  def _assimilate_table(self, mt):
    for key, val in mt._table.items():
      self._table[key] = val

  def __init__(self, *dicts, **kw):
    self._table = {}
    for d in list(dicts) + [kw]:
      if isinstance(d, dict):
        self._assimilate_dictionary(d)
      elif isinstance(d, Environment):
        self._assimilate_table(d)
      else:
        raise ValueError("Environment expects dict or Environment, got %s" % repr(d))

  # Duck-typed against provides classmethod.
  #
  # TODO(wickman) provides should probably somehow be integrated with find in this case.
  def provides(self, ref):
    assert isinstance(ref, Ref)
    if ref in self._table:
      return True
    targets = [key for key in self._table if Ref.subscope(key, ref)]
    if not targets:
      return False
    else:
      for key in sorted(targets, reverse=True):
        scope = self._table[key]
        if not isinstance(scope, Namable):
          continue
        subscope = Ref.subscope(key, ref)
        # If subscope is empty, then we should've found it in the ref table.
        assert not subscope.is_empty()
        if scope.provides(subscope):
          return True
    return False

  def find(self, ref):
    if ref in self._table:
      return self._table[ref]
    targets = [key for key in self._table if Ref.subscope(key, ref)]
    if not targets:
      raise Namable.NotFound(self, ref)
    else:
      for key in sorted(targets, reverse=True):
        scope = self._table[key]
        if not isinstance(scope, Namable):
          continue
        subscope = Ref.subscope(key, ref)
        # If subscope is empty, then we should've found it in the ref table.
        assert not subscope.is_empty()
        try:
          resolved = scope.find(subscope)
          return resolved
        except Namable.Error as e:
          continue
    raise Namable.NotFound(self, ref)

  def __repr__(self):
    return 'Environment(%s)' % pformat(self._table)


class Object(object):
  """
    Object base class, encapsulating a set of variable bindings scoped to this object.
  """

  class CoercionError(ValueError):
    def __init__(self, src, dst):
      ValueError.__init__(self, "Cannot coerce '%s' to %s" % (src, dst.__name__))

  class InterpolationError(Exception): pass

  @classmethod
  def checker(cls, obj):
    raise NotImplementedError

  def __init__(self):
    self._scopes = []
    self._modulo = TypeEnvironment()

  def get(self):
    raise NotImplementedError

  def __hash__(self):
    si, _ = self.interpolate()
    return hash(si.get())

  def copy(self):
    """
      Return a copy of this object.
    """
    raise NotImplementedError

  @staticmethod
  def translate_to_scopes(*args, **kw):
    scopes = []
    for arg in args:
      scopes.append(arg if isinstance(arg, Namable) else Environment.wrap(arg))
    scopes.extend([Environment(kw)] if kw else [])
    return scopes

  def bind(self, *args, **kw):
    """
      Bind environment variables into this object's scope.
    """
    new_self = self.copy()
    new_scopes = Object.translate_to_scopes(*args, **kw)
    new_self._scopes = list(reversed(new_scopes)) + new_self._scopes
    return new_self

  def provided(self, environment):
    new_self = self.copy()
    new_self._modulo = new_self._modulo.merge(environment)
    return new_self

  def in_scope(self, *args, **kw):
    """
      Scope this object to a parent environment (like bind but reversed.)
    """
    new_self = self.copy()
    new_scopes = Object.translate_to_scopes(*args, **kw)
    new_self._scopes = new_self._scopes + new_scopes
    return new_self

  def scopes(self):
    return self._scopes

  def modulo(self):
    return self._modulo

  def check(self):
    """
      Type check this object.
    """
    try:
      si, uninterp = self.interpolate()
    # TODO(wickman) This should probably be pushed out to the interpolate leaves.
    except (Object.CoercionError, MustacheParser.Uninterpolatable) as e:
      return TypeCheck(False, "Unable to interpolate: %s" % e)
    type_environment = self.modulo()
    for ref in uninterp:
      if not type_environment.covers(ref):
        return TypeCheck(False, "Uninterpolated variables: %s" %
          ' '.join('%s' % ref for ref in uninterp))
    return self.checker(si)

  def __ne__(self, other):
    return not (self == other)

  def __mod__(self, namable):
    if isinstance(namable, dict):
      namable = Environment.wrap(namable)
    interp, _ = self.bind(namable).interpolate()
    return interp

  def interpolate(self):
    """
      Interpolate this object in the context of the Object's environment.

      Should return a 2-tuple:
        The object with as much interpolated as possible.
        The remaining unbound Refs necessary to fully interpolate the object.

      If the object is fully interpolated, it should be typechecked prior to
      return.
    """
    raise NotImplementedError



import re

from pystachio.compatibility import Compatibility
from pystachio.naming import Namable, Ref

class MustacheParser(object):
  """
    Split strings on Mustache-style templates:
      a {{foo}} bar {{baz}} b => ['a ', Ref('foo'), ' bar ', Ref('baz'), ' b']

    To suppress parsing of individual tags, you can use {{&foo}} which emits '{{foo}}'
    instead of Ref('foo') or Ref('&foo').  As such, template variables cannot
    begin with '&'.
  """

  _ADDRESS_DELIMITER = '&'
  _MUSTACHE_RE = re.compile(r"{{(%c)?([^{}]+?)\1?}}" % _ADDRESS_DELIMITER)
  MAX_ITERATIONS = 100

  class Error(Exception): pass
  class Uninterpolatable(Error): pass

  @staticmethod
  def split(string, keep_aliases=False):
    splits = MustacheParser._MUSTACHE_RE.split(string)
    first_split = splits.pop(0)
    outsplits = [first_split] if first_split else []
    assert len(splits) % 3 == 0
    for k in range(0, len(splits), 3):
      if splits[k] == MustacheParser._ADDRESS_DELIMITER:
        outsplits.append('{{%s%s}}' % (
            MustacheParser._ADDRESS_DELIMITER if keep_aliases else '',
            splits[k+1]))
      elif splits[k] == None:
        outsplits.append(Ref.from_address(splits[k+1]))
      else:
        raise Exception("Unexpected parsing error in Mustache: splits[%s] = '%s'" % (
          k, splits[k]))
      if splits[k+2]:
        outsplits.append(splits[k+2])
    return outsplits

  @staticmethod
  def join(splits, *namables):
    """
      Interpolate strings.

      :params splits: The output of Parser.split(string)
      :params namables: A sequence of Namable objects in which the interpolation should take place.

      Returns 2-tuple containing:
        joined string, list of unbound object ids (potentially empty)
    """
    isplits = []
    unbound = []
    for ref in splits:
      if isinstance(ref, Ref):
        resolved = False
        for namable in namables:
          try:
            value = namable.find(ref)
            resolved = True
            break
          except Namable.Error as e:
            continue
        if resolved:
          isplits.append(value)
        else:
          isplits.append(ref)
          unbound.append(ref)
      else:
        isplits.append(ref)
    return (''.join(map(str if Compatibility.PY3 else unicode, isplits)), unbound)

  @classmethod
  def resolve(cls, stream, *namables):
    def iterate(st, keep_aliases=True):
      refs = MustacheParser.split(st, keep_aliases=keep_aliases)
      unbound = [ref for ref in refs if isinstance(ref, Ref)]
      repl, interps = MustacheParser.join(refs, *namables)
      rebound = [ref for ref in MustacheParser.split(repl, keep_aliases=keep_aliases)
                 if isinstance(ref, Ref)]
      return repl, interps, unbound

    iterations = 0
    for iteration in range(cls.MAX_ITERATIONS):
      stream, interps, unbound = iterate(stream, keep_aliases=True)
      if interps == unbound:
        break
    else:
      raise cls.Uninterpolatable('Unable to interpolate %s!  Maximum replacements reached.'
          % stream)

    stream, _, unbound = iterate(stream, keep_aliases=False)
    return stream, unbound



__author__ = 'Brian Wickman'
__version__ = '0.5.2'
__license__ = 'MIT'

import sys
if sys.version_info < (2, 6, 5):
  raise ImportError("pystachio requires Python >= 2.6.5")

from pystachio.typing import (
  Type,
  TypeCheck,
  TypeFactory)

from pystachio.base import Environment
from pystachio.parsing import MustacheParser
from pystachio.naming import Namable, Ref

from pystachio.basic import (
  Float,
  Integer,
  String)

from pystachio.container import (
  List,
  Map)

from pystachio.composite import (
  Default,
  Empty,
  Provided,
  Required,
  Struct)



import re
from itertools import chain

from pystachio.compatibility import Compatibility


class frozendict(dict):
  """A hashable dictionary."""
  def __key(self):
    return tuple((k, self[k]) for k in sorted(self))

  def __hash__(self):
    return hash(self.__key())

  def __eq__(self, other):
    return self.__key() == other.__key()

  def __repr__(self):
    return 'frozendict(%s)' % dict.__repr__(self)


class Namable(object):
  """
    An object that can be named/dereferenced.
  """
  class Error(Exception): pass

  class Unnamable(Error):
    def __init__(self, obj):
      Namable.Error.__init__(self, 'Object is not indexable: %s' % obj.__class__.__name__)

  class NamingError(Error):
    def __init__(self, obj, ref):
      Namable.Error.__init__(self, 'Cannot dereference object %s by %s' % (obj.__class__.__name__,
        ref.action()))

  class NotFound(Error):
    def __init__(self, obj, ref):
      Namable.Error.__init__(self, 'Could not find %s in object %s' % (ref.action().value,
        obj.__class__.__name__))

  def find(self, ref):
    """
      Given a ref, return the value referencing that ref.
      Raises Namable.NotFound if not found.
      Raises Namable.NamingError if try to dereference object in an invalid way.
      Raises Namable.Unnamable if try to dereference into an unnamable type.
    """
    raise NotImplementedError

  @classmethod
  def provides(cls, ref):
    """
      Given a ref, determine whether or not this class could dereference it.
    """
    raise NotImplementedError


class Ref(object):
  """
    A reference into to a hierarchically named object.
  """
  # ref re
  # ^[^\d\W]\w*\Z
  _DEREF_RE = r'[^\d\W]\w*'
  _INDEX_RE = r'[\w\-]+'
  _REF_RE = re.compile(r'(\.' + _DEREF_RE + r'|\[' + _INDEX_RE + r'\])')
  _VALID_START = re.compile(r'[a-zA-Z_]')
  _COMPONENT_SEPARATOR = '.'

  class Component(object):
    def __init__(self, value):
      self._value = value

    @property
    def value(self):
      return self._value

    def __eq__(self, other):
      return self.__class__ == other.__class__ and self.value == other.value

    def __ne__(self, other):
      return not (self == other)

    def __lt__(self, other):
      return self.value < other.value

    def __gt__(self, other):
      return self.value > other.value

  class Index(Component):
    RE = re.compile('^[\w\-]+$')
    def __repr__(self):
      return '[%s]' % self._value

  class Dereference(Component):
    RE = re.compile('^[^\d\W]\w*$')
    def __repr__(self):
      return '.%s' % self._value

  class InvalidRefError(Exception): pass
  class UnnamableError(Exception): pass

  @staticmethod
  def wrap(value):
    if isinstance(value, Ref):
      return value
    else:
      return Ref.from_address(value)

  @staticmethod
  def from_address(address):
    components = []
    if not address or not isinstance(address, Compatibility.stringy):
      raise Ref.InvalidRefError('Invalid address: %s' % repr(address))
    if not (address.startswith('[') or address.startswith('.')):
      if Ref._VALID_START.match(address[0]):
        components = Ref.split_components('.' + address)
      else:
        raise Ref.InvalidRefError(address)
    else:
      components = Ref.split_components(address)
    return Ref(components)

  def __init__(self, components):
    self._components = components

  def components(self):
    return self._components

  def action(self):
    return self._components[0]

  def is_index(self):
    return isinstance(self.action(), Ref.Index)

  def is_dereference(self):
    return isinstance(self.action(), Ref.Dereference)

  def is_empty(self):
    return len(self.components()) == 0

  def rest(self):
    return Ref(self.components()[1:])

  def __add__(self, other):
    sc = self.components()
    oc = other.components()
    return Ref(sc + oc)

  @staticmethod
  def subscope(ref1, ref2):
    rc = ref1.components()
    sc = ref2.components()
    if rc == sc[0:len(rc)]:
      if len(sc) > len(rc):
        return Ref(sc[len(rc):])

  def scoped_to(self, ref):
    return Ref.subscope(self, ref)

  @staticmethod
  def split_components(address):
    def map_to_namable(component):
      if (component.startswith('[') and component.endswith(']') and
          Ref.Index.RE.match(component[1:-1])):
        return Ref.Index(component[1:-1])
      elif component.startswith('.') and Ref.Dereference.RE.match(component[1:]):
        return Ref.Dereference(component[1:])
      else:
        raise Ref.InvalidRefError('Address %s has bad component %s' % (address, component))
    splits = Ref._REF_RE.split(address)
    if any(splits[0::2]):
      raise Ref.InvalidRefError('Badly formed address %s' % address)
    splits = splits[1::2]
    return [map_to_namable(spl) for spl in splits]

  def address(self):
    joined = ''.join(str(comp) for comp in self._components)
    if joined.startswith('.'):
      return joined[1:]
    else:
      return joined

  def __str__(self):
    return '{{%s}}' % self.address()

  def __repr__(self):
    return 'Ref(%s)' % self.address()

  def __eq__(self, other):
    return self.components() == other.components()

  @staticmethod
  def compare(self, other):
    if len(self.components()) < len(other.components()):
      return -1
    elif len(self.components()) > len(other.components()):
      return 1
    else:
      return (self.components() > other.components()) - (self.components() < other.components())

  def __lt__(self, other):
    return Ref.compare(self, other) == -1

  def __gt__(self, other):
    return Ref.compare(self, other) == 1

  def __hash__(self):
    return hash(str(self))



#!/usr/bin/env python
from setuptools import setup, find_packages, Command

version = '0.5.2'


class PyTest(Command):
  user_options = []

  def initialize_options(self):
    pass

  def finalize_options(self):
    pass

  def run(self):
    import sys, subprocess
    try:
      from py import test as pytest
    except ImportError:
      raise Exception('Running tests requires pytest.')
    errno = subprocess.call([sys.executable, '-m', 'py.test'])
    raise SystemExit(errno)


setup(
  name                 = 'pystachio',
  version              = version,
  description          = 'type-checked dictionary templating library',
  url                  = 'http://github.com/wickman/pystachio',
  author               = 'Brian Wickman',
  author_email         = 'wickman@gmail.com',
  license              = 'MIT',
  packages             = find_packages(),
  py_modules           = ['pystachio'],
  zip_safe             = True,
  cmdclass             = {
    'test': PyTest
  },
  scripts              = [
    'bin/pystachio_repl'
  ],
  classifiers          = [
    'Programming Language :: Python',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
  ],
)
